//
//	class CD3D11Sprite
//	Desc: This class implements 2D sprites on Direct3D11
//	
#include "stdafx.h"
#include <d3d11.h>
#include <directxmath.h>
#include <D3Dcompiler.h>
#include "D3D11Sprite.h"
#include "TextureLoader.h"
#include <vector>

//  
#define SAFE_DELETE(o)       if (o){  delete (o);  o = NULL;    }
#define SAFE_DELETE_ARRAY(o) if (o){  delete [] (o);  o = NULL; }
#define SAFE_RELEASE(o)      if (o){  (o)->Release(); o = NULL; }

static char shadercode[]="\
//-----------------------------------------------------------------------------------\n\
//  texture sampler\n\
//-----------------------------------------------------------------------------------\n\
SamplerState textureSampler : register(s0);\n\
Texture2D	diffuseTexture : register(t0);\n\
\n\
\n\
//-----------------------------------------------------------------------------------\n\
// Constant buffer\n\
//-----------------------------------------------------------------------------------\n\
cbuffer VSConstantBuffer : register( b0 )\n\
{\n\
    matrix  Proj;   //  ˉes\n\
	float4  Color;  //  W\gF\n\
};\n\
\n\
\n\
//-----------------------------------------------------------------------------------\n\
// VSInput structure\n\
//-----------------------------------------------------------------------------------\n\
struct VSInput\n\
{\n\
    float3 Position : POSITION;     //	ʒuW\n\
	float2 texCoord : TEXCOORD0;	//  eNX`W\n\
};\n\
\n\
//-----------------------------------------------------------------------------------\n\
// GSPSInput structure\n\
//-----------------------------------------------------------------------------------\n\
struct GSPSInput\n\
{\n\
    float4 Position : SV_POSITION;  //  ʒuW\n\
	float2 texCoord : TEXCOORD0;	//  eNX`W\n\
	float4 diffuse  : COLOR;        //  F\n\
};\n\
\n\
//-----------------------------------------------------------------------------------\n\
//	_VF[_Gg[|Cg\n\
//-----------------------------------------------------------------------------------\n\
GSPSInput VSFunc( VSInput input )\n\
{\n\
    GSPSInput output = (GSPSInput)0;\n\
\n\
    // ̓f[^float4 ֕ϊ.\n\
	float4 pos = float4( input.Position, 1.0f );\n\
\n\
	// ˉeԂɕϊ.\n\
    float4 projPos  = mul( Proj,  pos );\n\
\n\
	output.Position = projPos;\n\
	output.texCoord = input.texCoord;\n\
	output.diffuse  = Color;\n\
    return output;\n\
}\n\
\n\
//------------------------------------------------------------------------------------\n\
//	sNZVF[_Gg[|Cg\n\
//------------------------------------------------------------------------------------\n\
float4 PSFunc( GSPSInput psin ) : SV_TARGET0\n\
{\n\
	float4 pixel = diffuseTexture.Sample(textureSampler, psin.texCoord);\n\
    pixel = psin.diffuse * pixel;\n\
	return pixel;\n\
}\n\
float4 PSFuncNoTex( GSPSInput psin ) : SV_TARGET0\n\
{\n\
	return psin.diffuse;\n\
}\n\
\n";

//  eNX`Ǘp̃m[h\
struct SpriteTexture{
	TCHAR				*pFilename;
	ID3D11Texture2D		*pTexture;
	ID3D11ShaderResourceView	*pTextureShaderResourceView;
	DWORD				dwSrcWidth;
	DWORD				dwSrcHeight;

	SpriteTexture(){
		pFilename = NULL;
		pTexture = NULL;
		pTextureShaderResourceView = NULL;
	}
	~SpriteTexture(){
		SAFE_DELETE(pFilename);
		SAFE_RELEASE(pTexture);
		SAFE_RELEASE(pTextureShaderResourceView);
	}
};

//  eNX`Ǘpf[^
//  stl ̓WJǏ邽߃[Jf[^\ƂĂ
struct SpriteDataContext{
	std::vector<SpriteTexture *>	pSpriteTextures;
};

//  Vertex Structure
typedef struct {
	DirectX::XMFLOAT3 position;     //  ʒuW
	DirectX::XMFLOAT2 texture;      //  eNX`W
}	SpriteVertex;

//  Constant buffer structure
typedef struct {
	DirectX::XMMATRIX	matProj;
	DirectX::XMFLOAT4	color;
}	SpriteConstantBuffer;

//
//  constructor
//
CD3D11Sprite::CD3D11Sprite(void)
{
	InitFields();
	m_pDataContext = new SpriteDataContext;
}

//
//  destructor
//
CD3D11Sprite::~CD3D11Sprite(void)
{
	ReleaseDeviceObjects();
	std::vector<SpriteTexture *>::iterator it;
	it = m_pDataContext->pSpriteTextures.begin();
	while(it != m_pDataContext->pSpriteTextures.end()){
		SAFE_DELETE(*it);
		++it;
	}
	SAFE_DELETE(m_pDataContext);
}

//  StB[h̏
void CD3D11Sprite::InitFields(){

	m_pDataContext = NULL;
	m_pDeviceContext = NULL;
	m_pDevice = NULL;
	m_pTextureSamplerState = NULL;
	m_pVertexShader = NULL;
	m_pPixelShader = NULL;
	m_pPixelShaderNoTex = NULL;
	m_pInputLayout = NULL;
	m_pBlendState = NULL;
	m_pVertexBuffer = NULL;
	m_pConstantBuffer = NULL;

}

//
//	Restore device dependent objects.
//
HRESULT CD3D11Sprite::RestoreDeviceObjects(ID3D11DeviceContext *pContext){
	ID3DBlob	*pShader = NULL;
	HRESULT hr = E_FAIL;

	pContext->AddRef();
	ReleaseDeviceObjects();

	m_pDeviceContext = pContext;
	m_pDeviceContext->GetDevice(&m_pDevice);
	
	if (m_pDataContext){
		std::vector<SpriteTexture*>::iterator it;
		it = m_pDataContext->pSpriteTextures.begin();
		while (it != m_pDataContext->pSpriteTextures.end()){
			if (*it != NULL){
				ID3D11Texture2D		*pTexture = NULL;
				ID3D11ShaderResourceView	*pTextureShaderResourceView = NULL;
				
				hr = CTextureLoader::CreateTextureFromFile(pContext,
					(*it)->pFilename,   &pTexture,
					&(*it)->dwSrcWidth, &(*it)->dwSrcHeight,FillMode::None
				);
				// VF[_\[Xr[𐶐.
				if (SUCCEEDED(hr)){
					D3D11_SHADER_RESOURCE_VIEW_DESC srvd;
					D3D11_TEXTURE2D_DESC	texDesc;
					ZeroMemory( &srvd, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
					pTexture->GetDesc(&texDesc);
					srvd.Format                     = texDesc.Format;
  					srvd.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
					srvd.Texture2D.MipLevels = texDesc.MipLevels;
					srvd.Texture2D.MostDetailedMip = 0;
					hr = m_pDevice->CreateShaderResourceView( pTexture, &srvd, &pTextureShaderResourceView);
				}
				if (FAILED(hr)){
					SAFE_RELEASE(pTexture);
					SAFE_RELEASE(pTextureShaderResourceView);
				}
				(*it)->pTexture = pTexture;
				(*it)->pTextureShaderResourceView = pTextureShaderResourceView;
			}
			++it;
		}
	}

	//  _VF[_̐
	hr = CompileShaderFromMemory((BYTE*)shadercode,sizeof(shadercode),"VSFunc","vs_4_0",&pShader);
	if (FAILED(hr)){
		return	hr;
	}
	hr = m_pDevice->CreateVertexShader(pShader->GetBufferPointer(),pShader->GetBufferSize(),NULL,&m_pVertexShader);
	if (FAILED(hr)){
		return	hr;
	}


	// ̓CAEg̒`.
	D3D11_INPUT_ELEMENT_DESC layout[] = {
	    {
			"POSITION", 
			0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 
		},
	    {
			"TEXCOORD", 
			0, DXGI_FORMAT_R32G32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 
		}
	};
	UINT numElements = ARRAYSIZE( layout );

	// ̓CAEg𐶐.
	hr = m_pDevice->CreateInputLayout( 
	    layout,
	    numElements,
	    pShader->GetBufferPointer(),
	    pShader->GetBufferSize(),
	    &m_pInputLayout
	);
	SAFE_RELEASE(pShader);
	if ( FAILED( hr ) )
	{
		return hr;
	}

	// ̓AZuɓ̓CAEgݒ.
	m_pDeviceContext->IASetInputLayout( m_pInputLayout );

	//  sNZVF[_̐
	hr = CompileShaderFromMemory((BYTE*)shadercode,sizeof(shadercode),"PSFunc","ps_4_0",&pShader);
	if (FAILED(hr))
		return hr;
	hr = m_pDevice->CreatePixelShader(pShader->GetBufferPointer(),pShader->GetBufferSize(),NULL,&m_pPixelShader);
	SAFE_RELEASE(pShader);
	if (FAILED(hr))
		return	hr;
	//  sNZVF[_̐
	hr = CompileShaderFromMemory((BYTE*)shadercode,sizeof(shadercode),"PSFuncNoTex","ps_4_0",&pShader);
	if (FAILED(hr))
		return hr;
	hr = m_pDevice->CreatePixelShader(pShader->GetBufferPointer(),pShader->GetBufferSize(),NULL,&m_pPixelShaderNoTex);
	SAFE_RELEASE(pShader);
	if (FAILED(hr))
		return	hr;

	//  eNX`Tv\̃ZbgAbv
	{
		D3D11_SAMPLER_DESC samplerDesc;
		samplerDesc.Filter = D3D11_FILTER_ANISOTROPIC;         // TvOɎgptB^Bł͈ٕtB^[gpB
		samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;     // 0 ` 1 ͈̔͊Oɂ u eNX`[W̕`@
		samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;     // 0 ` 1 ͈̔͊Oɂ v eNX`[W
		samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;     // 0 ` 1 ͈̔͊Oɂ w eNX`[W
		samplerDesc.MipLODBias = 0;                            // vZꂽ~bv}bv x̃oCAX
		samplerDesc.MaxAnisotropy = 16;                        // TvOɈٕԂgpĂꍇ̌ElBLȒl 1 ` 16 B
		samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;  // rIvVB
		memcpy((void*)&samplerDesc.BorderColor,(void*)&DirectX::XMFLOAT4(0,0,0,0),4*sizeof(FLOAT));
		samplerDesc.MinLOD = 0;                                // ANZX\ȃ~bv}bv̉l
		samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;                // ANZX\ȃ~bv}bv̏l
		// ID3D11Device::CreateSamplerState
		hr = m_pDevice->CreateSamplerState( &samplerDesc, &m_pTextureSamplerState );
		if( FAILED( hr ) )
			return	hr;

	}
	// _obt@̐ݒ.
	{
		// _̒`.
		SpriteVertex vertices[] = {
			{  DirectX::XMFLOAT3( -0.5f,  0.5f, 0.0f ),  DirectX::XMFLOAT2(  0.0f,  0.0f ) },
			{  DirectX::XMFLOAT3(  0.5f,  0.5f, 0.0f ),  DirectX::XMFLOAT2(  1.0f,  0.0f ) },
			{  DirectX::XMFLOAT3( -0.5f, -0.5f, 0.0f ),  DirectX::XMFLOAT2(  0.0f,  1.0f ) },
			{  DirectX::XMFLOAT3(  0.5f, -0.5f, 0.0f ),  DirectX::XMFLOAT2(  1.0f,  1.0f ) },
		};
  
		D3D11_BUFFER_DESC bd;
		ZeroMemory( &bd, sizeof( D3D11_BUFFER_DESC ) );
		bd.Usage          = D3D11_USAGE_DEFAULT;
		bd.ByteWidth      = sizeof( SpriteVertex ) * 4;
		bd.BindFlags      = D3D11_BIND_VERTEX_BUFFER;
		bd.CPUAccessFlags = 0;

		// Tu\[X̐ݒ.
		D3D11_SUBRESOURCE_DATA initData;
		ZeroMemory( &initData, sizeof( D3D11_SUBRESOURCE_DATA ) );
		initData.pSysMem = vertices;

		// _obt@̐.
		hr = m_pDevice->CreateBuffer( &bd, &initData, &m_pVertexBuffer );
		if ( FAILED( hr ) )
		{
			return	hr;
		}
	}
	// 萔obt@̐.
	{
		// 萔obt@̐ݒ.
		D3D11_BUFFER_DESC bd;
		ZeroMemory( &bd, sizeof( D3D11_BUFFER_DESC ) );
		bd.ByteWidth        = sizeof( SpriteConstantBuffer );
		bd.Usage            = D3D11_USAGE_DEFAULT;
		bd.BindFlags        = D3D11_BIND_CONSTANT_BUFFER;
		bd.CPUAccessFlags   = 0;

		// 萔obt@𐶐.
		hr = m_pDevice->CreateBuffer( &bd, NULL, &m_pConstantBuffer );
		if ( FAILED( hr ) )
		{
			return	hr;
		}
	}
	{
		//  At@uhXe[g̐
		D3D11_BLEND_DESC BlendDesc;
		ZeroMemory( &BlendDesc, sizeof( BlendDesc ) );
		BlendDesc.AlphaToCoverageEnable = FALSE;
		BlendDesc.IndependentBlendEnable = FALSE;
		BlendDesc.RenderTarget[0].BlendEnable = TRUE;
		BlendDesc.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;
		BlendDesc.RenderTarget[0].DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
		BlendDesc.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
		BlendDesc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ONE;
		BlendDesc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
		BlendDesc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
		BlendDesc.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE_ALL;
		for (int i = 1 ; i < 8 ; ++i){
			BlendDesc.RenderTarget[i] = BlendDesc.RenderTarget[0];
		}
		hr = m_pDevice->CreateBlendState( &BlendDesc, &m_pBlendState );
		if ( FAILED( hr ) )
		{
			return	hr;
		}
	}
	// s̐ݒ.
	{
		//  non scaled ortho matrix (2D)
		m_matProj = DirectX::XMMatrixIdentity();
		m_matProj.r[0].m128_f32[0] = 1.0f;
		m_matProj.r[1].m128_f32[1] = 1.0f;
		m_matProj.r[2].m128_f32[2] = 0.0f;
		m_matProj.r[2].m128_f32[3] = 1.0f;
		m_matProj.r[3].m128_f32[2] = 0.0f;
		m_matProj.r[3].m128_f32[3] = 0.0f;
	}

	hr = S_OK;
	return hr;
}
//
//	Release device dependent objects.
//
HRESULT CD3D11Sprite::ReleaseDeviceObjects(){

	if (m_pDataContext){
		std::vector<SpriteTexture*>::iterator it;
		it = m_pDataContext->pSpriteTextures.begin();
		while (it != m_pDataContext->pSpriteTextures.end()){
			if (*it != NULL){
				SAFE_RELEASE((*it)->pTexture);
				SAFE_RELEASE((*it)->pTextureShaderResourceView);
			}
			++it;
		}
	}
	SAFE_RELEASE(m_pDeviceContext);
	SAFE_RELEASE(m_pDevice);
	SAFE_RELEASE(m_pTextureSamplerState);
	SAFE_RELEASE(m_pVertexShader);
	SAFE_RELEASE(m_pPixelShader);
	SAFE_RELEASE(m_pPixelShaderNoTex);
	SAFE_RELEASE(m_pInputLayout);
	SAFE_RELEASE(m_pBlendState);
	SAFE_RELEASE(m_pVertexBuffer);
	SAFE_RELEASE(m_pConstantBuffer);

	return S_OK;
}

//
//  method : SetTexture
//  @param:
//		no : texture no
//		pFilename : filename
//
void CD3D11Sprite::SetTexture(INT no, TCHAR *pFilename){
	int count = m_pDataContext->pSpriteTextures.size();
	int len;
	SpriteTexture *pTextureNode;
	if (no >= count){
		int add = count - no + 1;
		for ( int i = 0; i < add ; ++i){
			m_pDataContext->pSpriteTextures.push_back(NULL);
		}
	}
	pTextureNode = m_pDataContext->pSpriteTextures[no];
	if (pTextureNode == NULL){
		pTextureNode = new SpriteTexture;
		m_pDataContext->pSpriteTextures[no] = pTextureNode;
	}else{
		SAFE_DELETE_ARRAY(pTextureNode->pFilename);
		SAFE_RELEASE(pTextureNode->pTexture);
		SAFE_RELEASE(pTextureNode->pTextureShaderResourceView);
	}
	len = _tcslen(pFilename) + 1;
	len = (len + 3)&0xfffffffc;
	pTextureNode->pFilename = new TCHAR[len];
	_tcscpy_s(pTextureNode->pFilename,len,pFilename);
}

INT CD3D11Sprite::GetNumTextures(){
	INT num = 0;
	if (m_pDataContext){
		num = m_pDataContext->pSpriteTextures.size();
	}
	return num;
}

//
//  method : Render
//	@param : 
//		pContext : D3D11 context
//		x : position - x
//		y : position - y
//		w : sprite width
//		h : sprire height
//		no : texture no
//
void    CD3D11Sprite::Render(ID3D11DeviceContext *pContext, FLOAT x, FLOAT y, FLOAT w, FLOAT h, INT no){
	Render(pContext,x,y,w,h,0.0f,0.0f,FLT_MAX,FLT_MAX,no);
}

//
//  method : Render
//	@param : 
//		pContext : D3D11 context
//		x : position - x
//		y : position - y
//		w : sprite width
//		h : sprire height
//		tx : texture-coord.x
//		ty : texture-coord.y
//		tw : texture width
//		th : texture height
//		no : texture no
//
void    CD3D11Sprite::Render(ID3D11DeviceContext *pContext, FLOAT x, FLOAT y, FLOAT w, FLOAT h,
						FLOAT tx, FLOAT ty, FLOAT tw, FLOAT th, INT no){

	SpriteTexture			*pSpriteTexture = NULL;
	ID3D11Texture2D			*pTexture = NULL;
	ID3D11ShaderResourceView *pTextureShaderResourceView = NULL;

	if (m_pDataContext == NULL)
		return;

	if (no >= 0 && no < (INT)m_pDataContext->pSpriteTextures.size()){
		pSpriteTexture = m_pDataContext->pSpriteTextures[no];
		if (pSpriteTexture != NULL){
			pTexture = pSpriteTexture->pTexture;
			pTextureShaderResourceView = pSpriteTexture->pTextureShaderResourceView;
		}
	}
	if (pSpriteTexture == NULL)
		return;

	//@VF[_ݒ肵ĕ`.
	pContext->VSSetShader( m_pVertexShader, NULL, 0 );
	pContext->GSSetShader( NULL,			NULL, 0 );
	if (pTextureShaderResourceView == NULL){
		pContext->PSSetShader( m_pPixelShaderNoTex,  NULL, 0 );	//  eNX`
	}else{
		pContext->PSSetShader( m_pPixelShader,  NULL, 0 );		//  eNX`L
	}

	// 萔obt@̐ݒ.
	SpriteConstantBuffer cb;
	cb.matProj  = m_matProj;
	cb.color = m_vecDiffuse;

	// Tu\[XXV.
	pContext->UpdateSubresource( m_pConstantBuffer, 0, NULL, &cb, 0, 0 );

	// WIgVF[_ɒ萔obt@ݒ.
	pContext->VSSetConstantBuffers( 0, 1, &m_pConstantBuffer );

	float blendFactor[4] = {0.0f, 0.0f, 0.0f, 0.0f};
	pContext->OMSetBlendState( m_pBlendState, blendFactor, 0xffffffff );

	// _obt@̐ݒ.
	{
		D3D11_VIEWPORT  vp;
		UINT			uiNumViewport = 1;
		m_pDeviceContext->RSGetViewports(&uiNumViewport,&vp);
		if (uiNumViewport != 1)
			return;
		//  transform 2D -> clipping space
		FLOAT l = x;
		FLOAT t = y;
		FLOAT r = l + w;
		FLOAT b = t + h;
		FLOAT z = 1.0f;
		FLOAT tx1, ty1;
		FLOAT tx2, ty2;

		l /= vp.Width*0.5f;
		t /= vp.Height*0.5f;
		r /= vp.Width*0.5f;
		b /= vp.Height*0.5f;
		l -= 1.0f;
		r -= 1.0f;
		t = 1.0f - t;
		b = 1.0f - b;

		tx1 = ty1 = 0.0f;
		tx2 = ty2 = tx1;

		if (pTexture != NULL){
			D3D11_TEXTURE2D_DESC desc;
			FLOAT texW, texH;
			pTexture->GetDesc(&desc);
			texW = (FLOAT)desc.Width;
			texH = (FLOAT)desc.Height;
			tx1 = tx / texW;
			ty1 = ty / texH;
			if (tw == FLT_MAX){
				tx2 = (FLOAT)pSpriteTexture->dwSrcWidth / texW;
				ty2 = (FLOAT)pSpriteTexture->dwSrcHeight / texH;
			}else{
				tx2 = tx1 + tw / texW;
				ty2 = ty1 + th / texH;
			}
		}

		// _̒`.
		SpriteVertex vertices[] = {
			{  DirectX::XMFLOAT3( l, t, z ),  DirectX::XMFLOAT2(  tx1,  ty1 ) },
			{  DirectX::XMFLOAT3( r, t, z ),  DirectX::XMFLOAT2(  tx2,  ty1 ) },
			{  DirectX::XMFLOAT3( l, b, z ),  DirectX::XMFLOAT2(  tx1,  ty2 ) },
			{  DirectX::XMFLOAT3( r, b, z ),  DirectX::XMFLOAT2(  tx2,  ty2 ) },
		};
		pContext->UpdateSubresource( m_pVertexBuffer, 0, NULL, vertices, 0, 0 );
		
	}
	// ̓AZuɒ_obt@ݒ.
	UINT stride = sizeof( SpriteVertex );
	UINT offset = 0;
	pContext->IASetVertexBuffers( 0, 1, &m_pVertexBuffer, &stride, &offset );

	// v~eBu̎ނݒ.
	pContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );

	ID3D11ShaderResourceView* ppShaderResourceViews[] = { pTextureShaderResourceView, 0 };
	ID3D11SamplerState	*ppSamplerStates[] = { m_pTextureSamplerState, 0 };
	pContext->PSSetShaderResources(0, 1, ppShaderResourceViews);
	pContext->PSSetSamplers( 0, 1, ppSamplerStates );

	pContext->Draw(4,0);	//  Draw 4 vertices.

	ppSamplerStates[0] = NULL;
	ppShaderResourceViews[0] = NULL;
	pContext->PSSetSamplers( 0, 1, ppSamplerStates );
	pContext->PSSetShaderResources(0, 1, ppShaderResourceViews);

	//  uhXe[gɖ߂
	pContext->OMSetBlendState( NULL, blendFactor, 0xffffffff );
}

//
//	Set diffuse color.
//
void	CD3D11Sprite::SetDiffuse(FLOAT r, FLOAT g, FLOAT b, FLOAT a){
	m_vecDiffuse.x = r;
	m_vecDiffuse.y = g;
	m_vecDiffuse.z = b;
	m_vecDiffuse.w = a;
}

#define D3D_COMPILE_STANDARD_FILE_INCLUDE ((ID3DInclude*)(UINT_PTR)1)

HRESULT CD3D11Sprite::CompileShaderFromMemory(BYTE* pCode, DWORD dwSize, LPCSTR szEntryPoint, LPCSTR szShaderModel, ID3DBlob **ppBlobOut){
    // ^[R[h.
    HRESULT hr = S_OK;

    // RpCtO.
    DWORD dwShaderFlags = D3DCOMPILE_ENABLE_STRICTNESS;

#if defined(DEBUG) || defined(_DEBUG)
    dwShaderFlags |= D3DCOMPILE_DEBUG;
#endif//defiend(DEBUG) || defined(_DEBUG)

#if defined(NDEBUG) || defined(_NDEBUG)
    //dwShaderFlags |= D3DCOMPILE_OPTIMIZATION_LEVEL3;
#endif//defined(NDEBUG) || defined(_NDEBUG)

    ID3DBlob* pErrorBlob = NULL;

	// VF[_RpC.
	hr = D3DCompile(pCode,dwSize,NULL,NULL,D3D_COMPILE_STANDARD_FILE_INCLUDE,
		szEntryPoint,szShaderModel,dwShaderFlags,0,ppBlobOut,&pErrorBlob);

	// G[`FbN.
    if ( FAILED( hr ) )
    {
        // G[bZ[Wo.
        if ( pErrorBlob != NULL )
        { OutputDebugStringA( (char*)pErrorBlob->GetBufferPointer() ); }
    }

    // .
    if ( pErrorBlob )
    {
        pErrorBlob->Release();
        pErrorBlob = NULL;
    }

    // ^[R[hԋp.
	return hr;
}
